@@ -1,10 +1,12 @@
 # coding=utf-8
-from functools import wraps
+
 from io import StringIO
 import requests
-from .utils import *
+import pandas as pd
+from .utils import *  # noqa
 from .client import JQDataClient
 
+
 @assert_auth
 def get_price(security, start_date=None, end_date=None, frequency='daily',
               fields=None, skip_paused=False, fq='pre', count=None, panel=True, fill_paused=True):
@@ -25,7 +27,7 @@ def get_price(security, start_date=None, end_date=None, frequency='daily',
     :param fill_paused : False 表示使用NAN填充停牌的数据，True表示用close价格填充，默认True
     :return 如果是一支证券, 则返回pandas.DataFrame对象, 行索引是datetime.datetime对象, 列索引是行情字段名字; 如果是多支证券, 则返回pandas.Panel对象, 里面是很多pandas.DataFrame对象, 索引是行情字段(open/close/…), 每个pandas.DataFrame的行索引是datetime.datetime对象, 列索引是证券代号.
     """
-    if check_pandas_version():
+    if panel and PandasChecker.check_version():
         panel = False
     security = convert_security(security)
     start_date = to_date_str(start_date)
@@ -74,10 +76,16 @@ def get_fundamentals(query_object, date=None, statDate=None):
     查询财务数据, 详细的数据字段描述在 https://www.joinquant.com/data/dict/fundamentals 中查看
 
     :param query_object 一个sqlalchemy.orm.query.Query对象
-    :param date 查询日期, 一个字符串(格式类似’2015-10-15’)或者datetime.date/datetime.datetime对象, 可以是None, 使用默认日期
-    :param statDate: 财报统计的季度或者年份, 一个字符串, 有两种格式:1.季度: 格式是: 年 + ‘q’ + 季度序号, 例如: ‘2015q1’, ‘2013q4’. 2.年份: 格式就是年份的数字, 例如: ‘2015’, ‘2016’.
-    :return 返回一个 pandas.DataFrame, 每一行对应数据库返回的每一行(可能是几个表的联合查询结果的一行), 列索引是你查询的所有字段;为了防止返回数据量过大, 我们每次最多返回10000行;当相关股票上市前、退市后，财务数据返回各字段为空
+    :param date 查询日期, 一个字符串(格式类似’2015-10-15’)或者datetime.date/datetime.datetime对象,
+                可以是None, 使用默认日期
+    :param statDate: 财报统计的季度或者年份, 一个字符串, 有两种格式:1.季度: 格式是: 年 + ‘q’ + 季度序号,
+                     例如: ‘2015q1’, ‘2013q4’. 2.年份: 格式就是年份的数字, 例如: ‘2015’, ‘2016’.
+    :return 返回一个 pandas.DataFrame, 每一行对应数据库返回的每一行(可能是几个表的联合查询结果的一行),
+            列索引是你查询的所有字段;为了防止返回数据量过大, 我们每次最多返回10000行;
+            当相关股票上市前、退市后，财务数据返回各字段为空
     """
+    if date:
+        date = to_date(date)
     if date is None and statDate is None:
         date = datetime.date.today() - datetime.timedelta(days=1)
 
@@ -110,7 +118,7 @@ def get_fundamentals_continuously(query_object, end_date=None, count=1, panel=Tr
     from .finance_service import fundamentals_redundant_continuously_query_to_sql
     from .calendar_service import CalendarService
 
-    if check_pandas_version():
+    if panel and PandasChecker.check_version():
         panel = False
 
     trade_days = CalendarService.get_trade_days(end_date=end_date, count=count)
@@ -129,6 +137,80 @@ def get_fundamentals_continuously(query_object, end_date=None, count=1, panel=Tr
         df.reset_index(drop=True, inplace=True)
         return df
 
+@assert_auth
+@hashable_lru(maxsize=3)
+def get_valuation(security_list, start_date=None, end_date=None, fields=None, count=None):
+    """ 获取多个标的在指定交易日范围内的市值表数据
+    Args:
+        security_list: 标的code字符串列表或者单个标的字符串
+        end_date: 查询结束时间
+        start_date: 查询开始时间，不能与count共用
+        count: 表示往前查询每一个标的count个交易日的数据，如果期间标的停牌，则该标的返回的市值数据数量小于count
+        fields: 财务数据中市值表的字段，返回结果中总会包含code、day字段，可用字段如下：
+            code	股票代码	带后缀.XSHE/.XSHG
+            day	日期	取数据的日期
+            capitalization	总股本(万股)
+            circulating_cap	流通股本(万股)
+            market_cap	总市值(亿元)
+            circulating_market_cap	流通市值(亿元)
+            turnover_ratio	换手率(%)
+            pe_ratio	市盈率(PE, TTM)
+            pe_ratio_lyr	市盈率(PE)s
+            pb_ratio	市净率(PB)
+            ps_ratio	市销率(PS, TTM)
+            pcf_ratio	市现率(PCF, 现金净流量TTM)
+    Returns:
+        返回一个dataframe，索引默认是pandas的整数索引，返回的结果中总会包含code、day字段。
+    """
+    start_date = to_date_str(start_date)
+    end_date = to_date_str(end_date)
+    security_list = convert_security(security_list)
+    return JQDataClient.instance().get_valuation(**locals())
+
+
+@assert_auth
+def get_history_fundamentals(security, fields, watch_date=None, stat_date=None,
+                             count=1, interval='1q', stat_by_year=False):
+    """ 获取多个季度/年度的三大财务报表和财务指标数据. 可指定单季度数据, 也可以指定年度数据.
+            可以指定观察日期, 也可以指定最后一个报告期的结束日期.
+    Args:
+        security: 股票代码或者股票代码列表。
+        fields: 要查询的财务数据的列, 季度数据和年度数据可选择的列不同. 示例:
+            [
+                balance.cash_equivalents,
+                cash_flow.net_deposit_increase,
+                income.total_operating_revenue,
+            ]
+        watch_date: 观察日期, 如果指定, 将返回 watch_date 日期前(包含该日期)发布的报表数据
+        stat_date: 统计日期, 可以是 '2019'/'2019q1'/'2018q4' 等格式, 如果指定, 将返回 stat_date 对应报告期及之前的历史报告期的报表数据,
+            注意: watch_date 和 stat_date 只能指定一个, 而且必须指定一个
+                如果没有 stat_date 指定报告期的数据, 则该数据会缺失一行.
+        count: 查询历史的多个报告期时, 指定的报告期数量. 如果股票历史报告期的数量小于 count, 则该股票返回的数据行数将小于 count
+        interval: 查询多个报告期数据时, 指定报告期间隔, 可选值: '1q'/'1y', 表示间隔一季度或者一年, 举例说明:
+            stat_date='2019q1', interval='1q', count=4, 将返回 2018q2,2018q3,2018q4,2019q1 的数据
+            stat_date='2019q1', interval='1y', count=4, 将返回 2016q1,2017q1,2018q1,2019q1 的数据
+            stat_by_year=True, stat_date='2018', interval='1y', count=4 将返回 2015/2016/2017/2018 年度的年报数据
+        stat_by_year: bool, 是否返回年度数据. 默认返回的按季度统计的数据(比如income表中只有单个季度的利润).
+            如果返回年度数据:
+                interval必须是 '1y'
+                如果指定了 stat_date 的话, stat_date 必须是一个数字, 表明统计的年份
+            stat_by_year 是 True 时, fields 可以选择 balance/income/cash_flow/indicator/bank_indicator/security_indicator/insurance_indicator 表中的列
+                        是 False 时, fields 可以选择 balance/income/cash_flow/indicator 表中的列
+    Returns:
+        pandas.DataFrame, 数据库查询结果. 数据格式同 get_fundamentals. 每个股票每个报告期(一季度或者一年)的数据占用一行.
+    Others:
+        推荐用户对结果使用df.groupby来进行分组分析数据
+    """
+    assert security, "security is required"
+    assert fields, "fields is required"
+    assert (not watch_date) or (not stat_date), "watch_date与stat_date有且只能有一个不是None"
+    assert watch_date or stat_date, "watch_date与stat_date有且只能有一个不是None"
+
+    watch_date = to_date_str(watch_date)
+    security = convert_security(security)
+    fields = convert_fields_to_str(fields)
+    return JQDataClient.instance().get_history_fundamentals(**locals())
+
 
 @assert_auth
 @hashable_lru(maxsize=3)
@@ -279,6 +361,23 @@ def get_security_info(code, date=None):
         return Security(**result)
 
 
+@assert_auth
+@hashable_lru(maxsize=3)
+def get_security_info2(code, date=None):
+    """
+    获取标的信息
+
+    :param code 证券代码
+    :param date 查询日期: 日期字符串/date对象/datetime对象, 注意传入datetime对象时忽略日内时间
+    :return Security2
+    """
+    assert code, "code is required"
+    date = to_date_str(date)
+    result = JQDataClient.instance().get_security_info(**locals())
+    if result:
+        return Security2(**result)
+
+
 @assert_auth
 @hashable_lru(maxsize=3)
 def get_all_trade_days():
@@ -385,15 +484,21 @@ def get_future_contracts(underlying_symbol, date=None):
 
 
 @assert_auth
-def get_dominant_future(underlying_symbol, date=None):
+def get_dominant_future(underlying_symbol, date=None, end_date=None):
     """
     获取主力合约对应的标的
 
     :param underlying_symbol 期货合约品种，如 ‘AG’(白银)
-    :return 主力合约对应的期货合约
+    :param date 日期（默认为当前时刻，当指定 end_date 时则表示开始日期）
+    :param end_date 结束日期，当指定该参数时表示获取一段时间的主力合约
+    :return 主力合约对应的期货合约（指定 end_date 时返回 pandas.Series，否则返回字符串）
     """
     dt = to_date_str(date)
-    return JQDataClient.instance().get_dominant_future(underlying_symbol=underlying_symbol, dt=dt)
+    cli = JQDataClient.instance()
+    if not end_date:
+        return cli.get_dominant_future(underlying_symbol=underlying_symbol, dt=dt)
+    end_dt = to_date_str(end_date)
+    return cli.get_dominant_future(underlying_symbol=underlying_symbol, dt=dt, end_dt=end_dt)
 
 
 @assert_auth
@@ -648,6 +753,23 @@ def get_fund_info(security, date=None):
     return JQDataClient.instance().get_fund_info(**locals())
 
 
+@assert_auth
+@hashable_lru(maxsize=3)
+def get_futures_info(securities=None, fields=('contract_multiplier',
+                                              'tick_size',
+                                              'trade_time')):
+    """
+    期货信息数据接口
+
+    :param securities: 一个或多个标的
+    :param fields: 期货属性字段：['contract_multiplier', 'tick_size', 'trade_time']
+    :return: {}
+    """
+    assert securities, "securities is required"
+    securities = convert_security(securities)
+    return JQDataClient.instance().get_futures_info(**locals())
+
+
 @assert_auth
 def get_query_count(field=None):
     """
@@ -730,25 +852,15 @@ def get_factor_effect(security, start_date, end_date, period, factor, group_num=
     assert period[-1] in ["D", "W", "M"], "period must be end with one of (\"D\", \"W\", \"M\")"
     return JQDataClient.instance().get_factor_effect(**locals())
 
-@assert_auth
-def get_data(api_name, **kwargs):
-    """通用接口
-    获取非run_query, alpha因子, technical_analysis查询的数据
-    :param apis_name 接口名称
-    :param kwargs 传入参数，且必须是关键字参数
-    """
-    assert api_name
-    return JQDataClient.instance().get_data(api_name=api_name, args=kwargs)
-
 
 @assert_auth
 @hashable_lru(maxsize=3)
 def get_all_factors():
     return JQDataClient.instance().get_all_factors(**locals())
 
+
 @assert_auth
-@hashable_lru(maxsize=3)
-def get_call_auction(security, start_date=None, end_date=None, fields=None):
+def get_call_auction(security, start_date, end_date, fields=None):
     """ 获取指定时间区间内集合竞价时的tick数据
     Args:
         security: 标的代码或者标的代码组成的列表
@@ -777,35 +889,191 @@ def get_call_auction(security, start_date=None, end_date=None, fields=None):
             get_call_auction('000001.XSHE','2019-08-10','2019-08-12')
         2. start_date和end_date不能为None，否则将抛出异常
     """
-    return JQDataClient.instance().get_call_auction(**locals())
+    if str(end_date) >= str(datetime.datetime.today()):
+        if datetime.datetime.now().time() < datetime.time(9, 30):
+            # 当天9:30之前数据可能变化, 不用缓存
+            return JQDataClient.instance().get_call_auction(**locals())
+    return exec_call_auction(security, start_date=start_date, end_date=end_date, fields=fields)
+
+
+@hashable_lru(maxsize=3)
+def exec_call_auction(security,start_date=None, end_date=None, fields=None):
+    return JQDataClient.instance().get_call_auction(security=security,
+                                                    start_date=start_date,
+                                                    end_date=end_date,
+                                                    fields=fields)
+
+
+@assert_auth
+def get_factor_kanban_values(universe=None, bt_cycle=None, category=None, model='long_only', **kwargs):
+    """获取因子面板数据
+    Args:
+        universe: 股票池
+            'hs300': 沪深300
+            'zz500': 中证500
+            'zz800': 中证800
+            'zz1000': 中证1000
+            'zzqz': 中证全指
+        bt_cycle: 测试周期
+            'month_3':近三个月
+            'year_1': 近一年
+            'year_3': 近三年
+            'year_10': 近十年
+        category: 分类
+            'basics': 基础类
+            'quality': 质量类
+            'pershare': 每股类
+            'emotion': 情绪类
+            'growth': 成长类
+            'risk': 风险类
+            'barra': 风险因子 - 风格因子
+            'technical': 技术类
+            'momentum': 动量类
+        model: 组合构建模型
+            'long_only': 纯多头组合
+            'long_short': 多空组合
+        **kwargs: 其他约束条件详见
+            https://www.joinquant.com/help/api/help?name=api#get_factor_kanban_values
+
+    """
+    assert  model in ('long_only', 'long_short'),\
+            "model 的值应为 'long_only', 'long_short' 中的一个"
+    return JQDataClient.instance().get_factor_kanban_values(
+            universe=universe,
+            bt_cycle=bt_cycle,
+            category=category,
+            model=model,
+            **kwargs
+        )
+
+
+@assert_auth
+def get_factor_style_returns(factors=None, start_date=None,
+                             end_date=None, count=None,
+                             universe=None, industry='sw_l1'):
+    """获取风格因子暴露收益率
+
+    参数：
+        factors : 因子名称，单个因子（字符串）或一个因子列表,支持风格因子如"size"及申万/聚宽一级行业如 "801010"和"HY001"
+        start_date : 开始日期，字符串或 datetime 对象
+        end_date : 结束日期，字符串或 datetime 对象，可以与 start_date 或 count 配合使用
+        count: 截止 end_date 之前交易日的数量（含 end_date 当日）
+        universe : 市场范围,默认为None代表全市场, 可选 : 'hs300': 沪深300 ; 'zz500': 中证500'; zz800': 中证800; 'zz1000':中证1000; 'zzqz':中证全指
+        industry : 行业选取,默认为'sw_l1', 可选 : 'sw_l1':申万一级,'jq_l1':聚宽一级
+
+    返回：
+        一个 DataFrame, index 是日期, columns为因子名, 值为因子暴露收益率
+    """
+    if count and start_date:
+        raise ParamsError("(start_date, count) only one param is required")
+    start_date = to_date_str(start_date)
+    end_date = to_date_str(end_date)
+    return JQDataClient.instance().get_factor_style_returns(**locals())
+
+
+@assert_auth
+def get_factor_specific_returns(security, start_date=None, end_date=None, count=None):
+    """获取风格因子的特异收益率
+
+    参数:
+        security : 股票代码, 或者股票代码组成的list
+        start_date : 开始日期，字符串或 datetime 对象
+        end_date : 结束日期，字符串或 datetime 对象，可以与 start_date 或 count 配合使用
+        count: 截止 end_date 之前交易日的数量（含 end_date 当日）
+
+    返回:
+        个股在风格因子上的特异收益率
+    """
+    if count and start_date:
+        raise ParamsError("(start_date, count) only one param is required")
+    start_date = to_date_str(start_date)
+    end_date = to_date_str(end_date)
+    return JQDataClient.instance().get_factor_specific_returns(**locals())
+
+
+@assert_auth
+def get_factor_stats(factor_names=None, universe_type='hs300',
+                     start_date=None, end_date=None, count=None,
+                     skip_paused=False, commision_fee=0.0):
+    """ 获取历史收益(多头)
+    参数：
+        factor_names : 因子名称，单个因子（字符串）或一个因子列表
+        universe_type : 股票池包括以下五种：
+                                        'hs300': 沪深300
+                                        'zz500': 中证500
+                                        'zz800': 中证800
+                                        'zz1000': 中证1000
+                                        'zzqz': 中证全指
+                        默认为'hs300'
+        start_date : 开始日期，字符串或 datetime 对象
+        end_date : 结束日期，字符串或 datetime 对象，可以与 start_date 或 count 配合使用
+        count: 截止 end_date 之前交易日的数量（含 end_date 当日）
+        skip_paused : 是否跳过停牌，bool，默认为False
+        commision_fee : 手续费，float，0.0/0.0008/0.0018, 默认为0.0
+
+    返回：
+        一个 dict，其中 key 是因子名称，value 是一个 pandas.DataFrame
+        DataFrame 的 index 是日期，column 1,2,3,4,5是一分位、二分位、三分位、四分位、五分位累积收益
+    """
+    if count and start_date:
+        raise ParamsError("(start_date, count) only one param is required")
+    assert universe_type in ('zz500', 'zz1000', 'hs300', 'zz800', 'zzqz'),\
+            "股票池应为 'zz500', 'zz1000', 'hs300', 'zz800', 'zzqz'中一个"
+    assert commision_fee in (0.0, 0.0008, 0.0018), "手续费应为 0.0, 0.0008, 0.0018 中一个"
+    start_date = to_date_str(start_date)
+    end_date = to_date_str(end_date)
+    return JQDataClient.instance().get_factor_stats(**locals())
+
+
+@assert_auth
+def get_data(api_name, **kwargs):
+    """通用数据获取接口
+
+    获取非 run_query, alpha 因子, technical_analysis 查询的数据
+    :param apis_name 接口名称
+    :param kwargs 传入参数，且必须是关键字参数
+    """
+    assert api_name
+    return JQDataClient.instance().get_data(api_name=api_name, args=kwargs)
+
+
+def get_privilege():
+    """获取当前拥有的权限"""
+    return JQDataClient.instance().get_privilege()
+
+
+def get_now_time():
+    """获取服务器当前时间戳"""
+    return JQDataClient.instance().get_now_time()
+
+
+def get_server_version():
+    """获取服务器版本"""
+    try:
+        version = JQDataClient.instance().get_server_version()
+    except Exception:
+        version = "1.0"
+    return version
+
+
+def get_test():
+    """测试连通性"""
+    return JQDataClient.instance().get_test()
+
 
 def read_file(path):
-    """
-    读取文件
-    """
+    """读取文件"""
     with open(path, 'rb') as f:
         return f.read()
 
 
 def write_file(path, content, append=False):
-    """
-    写入文件
-    """
+    """写入文件"""
     if isinstance(content, six.text_type):
         content = content.encode('utf-8')
     with open(path, 'ab' if append else 'wb') as f:
         return f.write(content)
 
 
-__all__ = ["normalize_code", "attribute_history_engine", "history_engine",]
-
-def _collect_func():
-    funcs = []
-    for func in globals().keys():
-        if func.startswith("get"):
-            funcs.append(func)
-    return funcs
-
-__all__.extend(_collect_func())
-
-del _collect_func
+__all__ = ["normalize_code", "attribute_history_engine", "history_engine"]
+__all__.extend([name for name in globals().keys() if name.startswith("get")])
