@@ -1,21 +1,34 @@
 # coding=utf-8
-import six
+
+import sys
+import re
 import json
 import copy
 import datetime
 from functools import wraps
 from collections import namedtuple
-import re
-import pandas as pd
-if six.PY2:
-    import cPickle as pickle
-else:
-    import pickle as pickle
+from importlib import import_module
 
-try:
-    from functools import lru_cache
-except ImportError:
-    from fastcache import lru_cache
+import six
+
+
+for _modname in ("functools", "fastcache", "functools32"):
+    try:
+        lru_cache = import_module(_modname).lru_cache
+    except (ImportError, AttributeError):
+        continue
+    else:
+        break
+else:
+    def lru_cache(*args, **kwargs):
+        def wrapper(func):
+            @wraps(func)
+            def _func(*args, **kwargs):
+                return func(*args, **kwargs)
+            _func.cache_info = lambda: None
+            _func.cache_clear = lambda: None
+            return _func
+        return wrapper
 
 Serialized = namedtuple('Serialized', 'json')
 
@@ -55,6 +68,30 @@ class Security(object):
         return self.code
 
 
+class Security2(Security):
+
+    __slots__ = ("code", "display_name", "name", "start_date", "end_date",
+                 "type", "parent",)
+
+    def __repr__(self):
+        if sys.version_info[0] < 3:
+            display_name = self.display_name.encode('utf-8')
+        else:
+            display_name = self.display_name
+        return (
+            "Security(code='{}', type='{}', start_date='{}', end_date='{}', "
+            "display_name='{}')"
+        ).format(
+            self.code, self.type, self.start_date, self.end_date, display_name
+        )
+
+    def __str__(self):
+        return self.code
+
+    def to_dict(self):
+        return {name: getattr(self, name, None) for name in self.__slots__}
+
+
 def no_sa_warnings(f):
     @wraps(f)
     def wrapper(*args, **kwds):
@@ -67,9 +104,15 @@ def no_sa_warnings(f):
 
 
 def get_tables_from_sql(sql):
-    """ 从 sql 语句中拿到所有引用的表名字 """
-    m = re.search(r'FROM (.*?)($| WHERE| GROUP| HAVING| ORDER)', sql, flags=re.M)
-    return [t.strip() for t in m.group(1).strip().split(',')] if m else []
+    """从 sql 语句中拿到所有引用的表名字"""
+    sql = re.sub(r'\s+', ' ', sql).replace('`', '').replace('"', '')
+    pattern = re.compile(
+        r'(?:\b(?:from)|(?:join)\b)\s+(\w+(?:\s*,\s*\w+)*)\b',
+        flags=re.I
+    )
+    matchs = re.findall(pattern, sql)
+    tables = ' '.join(matchs).replace(',', ' ')
+    return {tab for tab in tables.split()}
 
 
 @no_sa_warnings
@@ -95,17 +138,46 @@ def remove_duplicated_tables(sql):
     return unique_sql
 
 
+def _get_sql_session():
+    from sqlalchemy.orm import scoped_session, sessionmaker
+    session = scoped_session(sessionmaker())
+    return session
+
+
+_sql_session = _get_sql_session()
+
+
+class SqlQuery(import_module("sqlalchemy.orm").Query):
+
+    limit_value = None
+    offset_value = None
+
+    def limit(self, limit):
+        self.limit_value = limit
+        return super(SqlQuery, self).limit(limit)
+
+    def offset(self, offset):
+        self.offset_value = offset
+        return super(SqlQuery, self).offset(offset)
+
+
+def query(*args, **kwargs):
+    return SqlQuery(args, **kwargs).with_session(_sql_session)
+
+
 def compile_query(query):
-    """ 把一个 sqlalchemy query object 编译成mysql风格的 sql 语句 """
-    from sqlalchemy.sql import compiler
+    """把一个 sqlalchemy query object 编译成mysql风格的 sql 语句"""
+    # from sqlalchemy.sql import compiler
     from sqlalchemy.dialects import mysql as mysql_dialetct
     from pymysql.converters import conversions, escape_item, encoders
 
     dialect = mysql_dialetct.dialect()
     statement = query.statement
-    comp = compiler.SQLCompiler(dialect, statement)
-    comp.compile()
-    enc = dialect.encoding
+    # comp = compiler.SQLCompiler(dialect, statement, literal_binds=True)
+    compile_kwargs = {"render_postcompile": True}
+    comp = statement.compile(dialect=dialect, compile_kwargs=compile_kwargs)
+    # comp.sql_compile()
+    # enc = dialect.encoding
     comp_params = comp.params
     params = []
     for k in comp.positiontup:
@@ -166,6 +238,14 @@ def convert_security(s):
         raise ParamsError("security's type should be Security or list")
 
 
+def convert_fields_to_str(s):
+    if isinstance(s, (list, tuple)):
+        res = [str(item) for item in s]
+        return res
+    else:
+        raise ParamsError("fields's type should be list or tuple")
+
+
 def normal_security_code(code):
     """
 
@@ -218,37 +298,27 @@ def to_date(date):
     raise ParamsError("type error")
 
 
-def _get_session():
-    from sqlalchemy.orm import scoped_session, sessionmaker
-    session = scoped_session(sessionmaker())
-    return session
-
-
-session = _get_session()
-
-
-def query(*args, **kwargs):
-    return session.query(*args, **kwargs)
-
-
 def assert_auth(func):
     @wraps(func)
     def _wrapper(*args, **kwargs):
         from .client import JQDataClient
         if not JQDataClient.instance():
-            print("run jqdatasdk.auth first")
+            raise Exception("Please run jqdatasdk.auth first")
         else:
             return func(*args, **kwargs)
     return _wrapper
 
+
 def get_mac_address():
     import uuid
     mac = uuid.UUID(int=uuid.getnode()).hex[-12:].upper()
     return '%s:%s:%s:%s:%s:%s' % (mac[0:2], mac[2:4], mac[4:6], mac[6:8],mac[8:10], mac[10:])
 
+
 def hashable_lru(maxsize=16):
     def hashable_cache_internal(func):
         cache = lru_cache(maxsize=maxsize)
+
         def deserialize(value):
             if isinstance(value, Serialized):
                 return json.loads(value.json)
@@ -281,6 +351,7 @@ def hashable_lru(maxsize=16):
 
     return hashable_cache_internal
 
+
 def get_security_type(security):
     exchange = security[-4:]
     code = security[:-5]
@@ -302,7 +373,7 @@ def get_security_type(security):
             elif code[0] == "1":
                 return "fund"
         else:
-            raise Exception("找不到标的%s" % security_code)
+            raise Exception("找不到标的%s" % security)
     else:
         if exchange in ("XSGE", "XDCE", "XZCE", "XINE", "CCFX"):
             if len(code) > 6:
@@ -310,14 +381,23 @@ def get_security_type(security):
             return "future"
     return 0
 
-def check_pandas_version():
-    if pd.__version__[:4] >= "0.25":
-        return True
-    return False
 
-def get_pandas_notice():
-    return (
-        "提示：当前环境 pandas 版本高于 0.25，"
-        "get_price 与 get_fundamentals_continuously 接口 panel 参数将固定为 False\n"
-        "注意：0.25 以上版本 pandas 不支持 panel，如使用该数据结构和相关函数请注意修改"
+class PandasChecker(object):
+
+    @staticmethod
+    def check_version():
+        if import_module("pandas").__version__[:4] >= "0.25":
+            return True
+        return False
+
+    VERSION_NOTICE_MESSAGE = (
+        "当前环境 pandas 版本高于 0.25，get_price 与 get_fundamentals_continuously "
+        "接口的 panel 参数将固定为 False（0.25 及以上版本的 pandas 不再支持 panel，"
+        "如使用该数据结构和相关函数请注意修改）"
     )
+
+
+def isatty(stream=None):
+    stream = stream or sys.stdout
+    _isatty = getattr(stream, 'isatty', None)
+    return _isatty and _isatty()
